from datetime import datetime
import csv
import requests
import pandas as pd
import numpy as np
import numpy.ma as ma
import metpy.calc as mpcalc
from metpy.units import units
import copy







"""
    SOUNDERPY MODEL REANALYSIS 'GET-DATA' FUNCTIONS  

    Purpose of module: 

     A collection of helper functions that users may find useful for processing 
     vertical profile data but are not necessary to use the basic functions of 
     SounderPy. May be called by other modules or explicitly called by the user.


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024
"""





#########################
# MODIFY SURFACE VALUES OF SOUNDING ARRAYS
#########################################################################
def modify_surface(clean_data, modify_sfc):
    """
    where,
    modify_sfc = {'T': 20, 'Td': 12, 'ws': 10, 'wd': 270}
    and units = degC, degC, kts, degrees
    """

    sounding_data = copy.deepcopy(clean_data)

    # Modify arrays based on the provided modifications
    if 'T' in modify_sfc:
        sounding_data['T'] = barnes_interp(modify_sfc['T'], clean_data['T'], clean_data['p'])*units.degC

    if 'Td' in modify_sfc:
        sounding_data['Td'] = barnes_interp(modify_sfc['Td'], clean_data['Td'], clean_data['p'])*units.degC

    if 'ws' in modify_sfc:
        if 'wd' in modify_sfc:
            new_u, new_v = mpcalc.wind_components(modify_sfc['ws'] * units.kts, modify_sfc['wd'] * units.deg)

            sounding_data['u'] = barnes_interp(new_u.m, clean_data['u'], clean_data['p']) * units.kts
            sounding_data['v'] = barnes_interp(new_v.m, clean_data['v'], clean_data['p']) * units.kts

    return sounding_data





#########################
# BARNES INTERP VERTICAL ARRAYS FOR SFC MODIFICATION
#########################################################################
def barnes_interp(new_val, xarr, yarr, kappa=0.5, num_points=5):
    """
    conduct barnes interp on a vertical array of T,Td,u, or v arrays

    - new_val: The new value for the surface point (index 0).
    - xarr: The original array of data (temperature, dewpoint, u, or v).
    - yarr: The pressure levels corresponding to the data.
    - kappa: Smoothing parameter for Barnes interpolation.
    - num_points: The number of lowest points to apply the interpolation to (default is 5).
    """

    xarr = np.array(xarr)
    pressure_levels = np.array(yarr)

    # number of points to interp through (lowest 5)
    num_points = min(num_points, len(xarr))

    # replace first val in xarr with new sfc val
    xarr[0] = new_val

    # interpolate the xarr with the new val against the yarr
    interpolated_data = np.copy(xarr)
    for i in range(1, num_points):
        influence = np.exp(-((yarr[i] - yarr[0]) ** 2) / (kappa * (yarr[num_points - 1] - yarr[0]) ** 2))
        interpolated_data[i] = influence * new_val + (1 - influence) * xarr[i]

    # return interp'd xarr
    return interpolated_data





#########################
# INTERPOLATE DATA
#########################################################################
def interp_data(variable, heights, step=100):
    '''
    Interpolate a 1D array of data (such as a temperature profile) over a given interval (step) based on a corresponding array of height values.

    :param variable: an array of data to be interpolated. Must be same length as height array.
    :type variable: arr, required
    :param heights: heights corresponding to the vertical profile used to interpolate. Must be same length as variable array.
    :type heights: arr, required
    :param step: the resolution of interpolation. Default is 100 (recommended value is 100)
    :type step: int, optional
    :return: interp_var, an array of interpolated data.
    :rtype: arr
    '''

    try:
        variable.units
        variable = variable.m
    except:
        variable = variable
    try:
        heights.units
        heights = heights.m
    except:
        heights = heights

    levels = np.arange(0, np.max(heights), step)
    varinterp = np.zeros(len(levels))
    for i in range(0, len(levels)):
        lower = np.where(heights - levels[i] <= 0, heights - levels[i], -np.inf).argmax()
        varinterp[i] = (((variable[lower + 1] - variable[lower]) / (heights[lower + 1] - heights[lower])) * (
                    levels[i] - heights[lower]) + variable[lower])
    return varinterp






#########################
# FIND NEAREST
#########################################################################
def find_nearest(array, value):
    """
    search through an array to find the index of the value nearest to a given value
    """
    array = np.asarray(array)
    nearest_idx = (np.abs(array - value)).argmin()
    return nearest_idx








#########################
# TXT MANIPULATION | mag and mag_round functions
#########################################################################
# make text prettier
# magnitude number text
def mag(param):
    if ma.is_masked(param):
        fixed = '---'
    else:
        try:
            fixed = int(param.m)
        except:
            try:
                fixed = int(param)
            except:
                fixed = param
    return fixed


# make text prettier
# magnitude and round number text
# kawrgs - 'param': the value to round; 'dec': the decimal places to round to;
# 'mag': bool, remove units (magintude of the value)
def mag_round(param, dec, mag=False):
    if ma.is_masked(param):
        fixed = '---'
    elif mag == True:
        fixed = np.round(param.m, dec)
    else:
        fixed = np.round(param, dec)
    return fixed






#########################
# GET SFC INDEX
#########################################################################
def get_sfc_index(height_arr):
    """
    Return a value of an index of an array who's value is closest to a define value.

    :param array: an array of data to be searched through
    :type array: arr, required
    :param heights: the value used to compare against the array of data
    :type heights: int or float, required
    :return: nearest_idx, index of the data array that corresponds with the nearest value to the given value
    :rtype: int
    """

    i = 0
    # Search the array for the sfc index
    while i < len(height_arr):
        if height_arr[i] >= 0:
            return i
        else:
            i += 1
    # Did not find a positive index
    return -1







####################
# MAKE SURFACE BASED
#########################################################################
def make_sfc_based(arr, sfc_val, sfc_index):
    """
    takes an array and a valid index in that array, then returns a copy of the
    array beginning at the provided index i.e., chops off below-ground values
    """

    # Initialize an empty numpy array as the modified array
    mod_arr = np.empty(len(arr) - sfc_index)
    # Insert the sfc and higher values into the modified array
    i = 0
    while i < len(mod_arr):
        mod_arr[i] = arr[i + sfc_index]
        i = i + 1
    # Inserts surface values
    mod_arr = np.insert(mod_arr, 0, sfc_val)
    return mod_arr








#########################
# MAKE SURFACE-BASED 3D
#########################################################################
def make_sfc_based_3D(arr, sfc_arr):
    '''
    takes a 3D array of mandatory level data and a 2D array of surface data,
    appends the surface data onto the mandatory level array, and returns a single array
    of both surface and mandatory level data
    '''

    mod_arr = np.zeros((np.shape(arr)[0] + 1, np.shape(arr)[1], np.shape(arr)[2]))
    for j in range(np.shape(arr)[1]):
        for k in range(np.shape(arr)[2]):
            mod_arr[0, j, k] = sfc_arr[j, k]
            mod_arr[1:, j, k] = arr[:, j, k]
    return mod_arr







#########################
# FILE CREATION FUNCTIONS
#########################################################################

def to_file(file_type, clean_data, filename=None):
    '''
    Create a file of 'cleaned' SounderPy data

   :param file_type: a `str` representing the file type you'd like to export data to.
   :type file_type: str, required
   :param clean_data: 'cleaned' SounderPy data `dict`
   :type clean_data: dict, required
   :param filename: the name you'd like to give the file
   :type filename: str, required
   :return: a file of SounderPy data.
    '''

    # set file name
    if filename is None:
        filename = f'sounderpy_data'
    else:
        filename = filename



    ####################################### CM1 #######################################
    if file_type == 'cm1':
        '''
        creates CM1 input sounding file for CM1 integration

        Derived from Kelton Halbert / Leigh Orf via github: 
        https://github.com/leighorf/LOFS-read/blob/master/bin/sndmod
        '''

        # create file
        outfile = open(filename, 'w')
        num_lines = len(list(clean_data.items())[0][1])
        delimiter = ''

        # use metpy to find parameters that CM1 likes
        clean_data['theta'] = mpcalc.potential_temperature(clean_data['p'], clean_data['T'])
        clean_data['relhm'] = mpcalc.relative_humidity_from_dewpoint(clean_data['T'], clean_data['Td'])
        clean_data['mixrt'] = mpcalc.mixing_ratio_from_relative_humidity(clean_data['p'], clean_data['T'],
                                                                         clean_data['relhm'])*1000

        # create the sfc values line
        top_line = (
                "%12s" % str(format(np.around(clean_data["p"][0].m, 6), "0.6f")) + delimiter + "\t" +
                "%12s" % str(format(np.around(clean_data["theta"][0].m, 6), "0.6f")) + delimiter + "\t" +
                "%12s" % str(format(np.around(clean_data["mixrt"][0].m, 6), "0.6f")) + "\n"
        )

        # write the sfc values line to the file
        outfile.write(top_line)

        # add data to lines
        for idx in range(num_lines):
            line_str = ""
            line_str += "%12s" % str(format(np.around(clean_data["z"][idx].m, 6), "0.6f")) + delimiter + str("\t")
            line_str += "%12s" % str(format(np.around(clean_data["theta"][idx].m, 6), "0.6f")) + delimiter + str("\t")
            line_str += "%12s" % str(format(np.around(clean_data["mixrt"][idx].m, 6), "0.6f")) + delimiter + str("\t")
            line_str += "%12s" % str(format(np.around(clean_data["u"][idx].m/1.94384, 6), "0.6f")) + delimiter + str("\t")
            line_str += "%12s" % str(format(np.around(clean_data["v"][idx].m/1.94384, 6), "0.6f")) + str("\n")
            outfile.write(line_str)

        outfile.close()



    ####################################### CSV #######################################
    elif file_type == 'csv':
        '''
        creates CSV file of sounding data
        '''

        # remove units from data
        no_units = {}
        for key in ['p', 'z', 'T', 'Td', 'u', 'v']:
            no_units[key] = clean_data[key].m
        # open and write to CSV
        with open(filename, "w") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(no_units.keys())
            writer.writerows(zip(*no_units.values()))




    ####################################### SHARPPY #######################################
    elif file_type == 'sharppy':
        '''
        creates NSHARP input sounding file for SharpPy integration

        Derived from Kelton Halbert / Leigh Orf via github: 
        https://github.com/leighorf/LOFS-read/blob/master/bin/sndmod
        '''

        outfile_file = open(filename, 'w')

        outfile_loc = ("****")

        dt = datetime(int(clean_data['site_info']['valid-time'][0]), int(clean_data['site_info']['valid-time'][1]),
                      int(clean_data['site_info']['valid-time'][2]), int(clean_data['site_info']['valid-time'][3][0:2]))

        outfile_file.write("%TITLE%\n")
        outfile_file.write("%s   %s\n" % (clean_data['site_info']['site-id'], dt.strftime("%y%m%d/%H%M")))
        outfile_file.write("   LEVEL       HGHT       TEMP       DWPT       WDIR       WSPD\n")
        outfile_file.write("-------------------------------------------------------------------\n")
        outfile_file.write("%RAW%\n")

        ws = mpcalc.wind_speed(clean_data['u'], clean_data['v'])
        wd = mpcalc.wind_direction(clean_data['u'], clean_data['v'])

        new_data = {
            'p': clean_data['p'],
            'z': clean_data['z'],
            'T': clean_data['T'],
            'Td': clean_data['Td'],
            'wd': wd,
            'ws': ws,
        }

        for idx in range(new_data['p'].shape[0]):
            string = ""
            for col in ['p', 'z', 'T', 'Td', 'wd', 'ws']:
                string += "%12.6f,  " % new_data[col][idx].m

            outfile_file.write(string[:-3] + "\n")
        outfile_file.write("%END%\n")
        outfile_file.close()

    ##########################################################################################################################################








#########################
# LAT-LON FINDER FUNCTION
#########################################################################

def get_latlon(station_type, station_id):
    '''
    Return a latitude-longitude float pair in a ``list``

    :param station_type: the station 'type' that corresponds with the given station ID
    :type station_type: str, required
    :param station_id: the station ID for the given station type
    :type station_id: str, required
    :return: lat/lon float pair
    :rtype: list
    '''

    station_id = str.upper(station_id)

    # DMS to decimal degrees
    def dms2dd_min(degrees, minutes, direction):
        dd = float(degrees) + float(minutes) / 60
        if direction == "S" or direction == "W":
            dd *= -1
        return dd

    def dms2dd(degrees, direction):
        dd = float(degrees)
        if direction == "S" or direction == "W":
            dd *= -1
        return dd



    ############################################### METAR #####################################################
    if station_type.casefold() == 'metar':
        '''
        takes a METAR site id such as 'KMBS', searches over 9000 station IDs  and returns
        a list including the lat/lon for the METAR site
        '''
        # get METAR stations list
        request = requests.get("https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/METAR-STATIONS.txt",
                               stream=True)
        stations = {}
        for line in request.iter_lines():
            data = line.decode("ascii")
            if data:
                if data[0] == "!" or len(data) != 83:
                    continue
                province = data[0:2]
                station = data[3:19].strip()
                icao = data[20:24].strip()
                lat = dms2dd_min(data[39:41], data[42:44], data[44:45])
                lon = dms2dd_min(data[47:50], data[51:53], data[53:54])
                altitude = int(data[55:59])
                country = data[81:83]

                if icao:
                    stations[icao] = {"name": station, "lat": lat, "lon": lon, "altitude": altitude, "country": country}
        try:
            latlon = [np.round(stations.get(station_id)['lat'], 2), np.round(stations.get(station_id)['lon'], 2)]
            return latlon
        except:
            raise ValueError(f"The station you requested ({station_id}) doesn't seem to exist\n" +
                             "TIP: most METAR IDs include a 'K' in front, such as 'KMOP'")



    ############################################### BUFKIT #####################################################
    elif station_type.casefold() == 'bufkit':
        '''
        takes a BUFKIT site id such as 'KMOP', searches over 1200 station IDs and returns
        a list including the lat/lon for the BUFKIT site
        '''
        # get BUFKIT stations list
        BUFKIT_STATIONS = pd.read_csv(
            f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/BUFKIT-STATIONS-MASTER.txt',
            skiprows=7, skipinitialspace=True)

        try:
            station = BUFKIT_STATIONS['ID'][
                np.where(BUFKIT_STATIONS['ID'].str.contains(station_id, na=False, case=True))[0]].values[0]
            lat = (BUFKIT_STATIONS[BUFKIT_STATIONS['ID'] == station]['LAT'].values[0])
            lon = (BUFKIT_STATIONS[BUFKIT_STATIONS['ID'] == station]['LON'].values[0])
            return [lat, lon]
        except:
            raise ValueError(f"The station you requested ({station_id}) doesn't seem to exist\n" +
                             "TIP: some IDs include a 'K' in front, such as 'KMOP', others are 3 digits, such a 'DTX'")



    ############################################### RAOB #####################################################
    elif station_type.casefold() == 'raob':

        '''
        takes a RAOB site id such as 'DTX', searches over 9000 station IDs  and returns
        a list including the lat/lon for the RAOB site
        '''
        # get RAOB stations list
        RAOB_STATIONS = pd.read_csv(
            f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/RAOB-STATIONS.txt',
            skiprows=7, skipinitialspace=True)

        # find lat-lon from stations list if it exists
        try:
            station = RAOB_STATIONS['ICAO'][
                np.where(RAOB_STATIONS['ICAO'].str.contains(station_id, na=False, case=True))[0]].values[0]
            lat = dms2dd(RAOB_STATIONS[RAOB_STATIONS['ICAO'] == station]['LAT'].values[0],
                         RAOB_STATIONS[RAOB_STATIONS['ICAO'] == station]['A'].values[0])
            lon = dms2dd(RAOB_STATIONS[RAOB_STATIONS['ICAO'] == station]['LON'].values[0],
                         RAOB_STATIONS[RAOB_STATIONS['ICAO'] == station]['B'].values[0])
            return [lat, lon]
        except:
            try:
                station = RAOB_STATIONS['WMO'][RAOB_STATIONS[RAOB_STATIONS['WMO'] == int(station_id)].index[0]]
                lat = dms2dd(RAOB_STATIONS[RAOB_STATIONS['WMO'] == station]['LAT'].values[0],
                             RAOB_STATIONS[RAOB_STATIONS['WMO'] == station]['A'].values[0])
                lon = dms2dd(RAOB_STATIONS[RAOB_STATIONS['WMO'] == station]['LON'].values[0],
                             RAOB_STATIONS[RAOB_STATIONS['WMO'] == station]['B'].values[0])
                return [lat, lon]
            except:
                raise ValueError(f"The station you requested ({station_id}) doesn't seem to exist")



    ############################################### IGRA #####################################################
    elif station_type.casefold() == 'igra':
        '''
        takes a IGRA2 site id such as 'GMM00010393', searches nearly 3000 station IDs  and returns
        a list including the lat/lon for the IGRA2 site
        '''

        # get IGRA stations list
        IGRA_STATIONS = pd.read_csv(
            f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/IGRA-STATIONS.txt',
            skiprows=7, skipinitialspace=True)

        # find lat-lon from stations list if it exists
        try:
            station = \
            IGRA_STATIONS['ID'][np.where(IGRA_STATIONS['ID'].str.contains(station_id, na=False, case=True))[0]].values[
                0]

            lat = np.round(IGRA_STATIONS[IGRA_STATIONS['ID'] == station]['LAT'].values[0], 2)
            lon = np.round(IGRA_STATIONS[IGRA_STATIONS['ID'] == station]['LON'].values[0], 2)
            return [lat, lon]
        except:
            raise ValueError(f"The station you requested ({station_id}) doesn't seem to exist")



    ############################################### BUOY #####################################################
    elif station_type.casefold() == 'buoy':
        '''
        takes a BUOY/CMAN site id such as '41001', searches through a number of station IDs and returns
        a list including the lat/lon for the BUOY/CMAN  site
        '''

        # get buoy stations list
        BUOY_STATIONS = pd.read_csv(
            'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/BUOY-STATIONS.txt',
            skiprows=7, skipinitialspace=True)

        # find lat-lon from stations list if it exists
        try:
            station = \
            BUOY_STATIONS['ID'][np.where(BUOY_STATIONS['ID'].str.contains(station_id, na=False, case=True))[0]].values[
                0]

            lat = dms2dd(BUOY_STATIONS[BUOY_STATIONS['ID'] == station]['LAT'].values[0],
                         BUOY_STATIONS[BUOY_STATIONS['ID'] == station]['A'].values[0])
            lon = dms2dd(BUOY_STATIONS[BUOY_STATIONS['ID'] == station]['LON'].values[0],
                         BUOY_STATIONS[BUOY_STATIONS['ID'] == station]['B'].values[0])
            return [lat, lon]
        except:
            raise ValueError(f"The station you requested ({station_id}) doesn't seem to exist\n" +
                             "TIP: buoy IDs typically look like this: '41001'")
    else:
        raise ValueError(
            f"Incorrect station_type argument. Valid station_type-s are 'metar', 'raob', 'igra', 'bufkit', 'buoy'")



import time
import warnings
from datetime import datetime

import pandas as pd
import numpy as np

from metpy.units import units

from siphon.simplewebservice.wyoming import WyomingUpperAir
from siphon.simplewebservice.igra2 import IGRAUpperAir

from .utils import get_latlon
from .calc import sounding_params







"""
    SOUNDERPY RAOB OBSERVATIONS 'GET-DATA' FUNCTIONS  

    Purpose of module: 

    House function for loading and parsing RAOB & IGRAv2 observation
    data. Functions here are referenced by sounderpy.py


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024
"""






#######################
# RAOB AND IRGAv2 OBSERVATIONS DATA
#########################################################################

def fetch_obs(station, year, month, day, hour, hush, clean_it):

    # record process time
    st = time.time()

    r"""
       Return a ``dict`` of 'cleaned up' observed profile data
    
       :param station: a three digit RAOB identifier (such as: 'DTX') or 11 digit IGRAv2 identifier (such as: 'GMM00010393')
       :type station: str, required
       :param year: launch year
       :type year: str, required
       :param month: launch month
       :type month: str, required
       :param day: launch day
       :type day: str, required
       :param hour: launch hour
       :type hour: str, required
       :param clean_it: whether to return the raw_data object or a clean_data dict.
       :type clean_it: bool, optional, default is `True`
       :return: clean_data, a dict of ready-to-use vertical profile data including pressure, height, temperature, dewpoint, u-wind, v-wind, & model information
       :rtype: dict
    """

    print(f'> OBSERVED DATA ACCESS FUNCTION\n  -----------------------------------')

    station = str.upper(station)
    # get station lists from SounderPy GitHub Repo
    RAOB_STATIONS = pd.read_csv(f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/RAOB-STATIONS.txt',
                                skiprows=7, skipinitialspace=True)
    IGRA_STATIONS = pd.read_csv(f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/IGRA-STATIONS.txt',
                                skiprows=7, skipinitialspace=True)

    got_data = False

    # set up siphon API call for raob data -- if station ID is found in RAOB_STATIONS, it is
    # a RAOB ID and siphon UW or ISU must be used to get data
    if len(station) == 11:
        search_for = 'igra'
    else:
        search_for = 'raob'






    ### RAOB OBSERVATIONS ###
    #########################################################################################################
    if search_for == 'raob':
        # try this process 10 times, sometimes requests fail due to temporary 404 errors
        for i in range(1, 11):
            try:
                # try UW data request
                df = WyomingUpperAir.request_data(datetime(int(year), int(month), int(day), int(hour)), station)
                got_data = True
                if got_data:
                    print(f'    > PROFILE FOUND: {station} on {month}/{day}/{year} at {hour}z | From UW')
                    break
            except:
                got_data = False
                pass

        # search through RAOB sites list with provided RAOB ID, first try ICAO ID, then WMO ID
        if got_data:
            if clean_it:
                try:
                    station = \
                    RAOB_STATIONS['ICAO'][np.where(RAOB_STATIONS['ICAO'].str.contains(station, na=False, case=True))[0]].values[
                        0].strip()
                    name_idx = 'ICAO'
                except:
                    try:
                        station = RAOB_STATIONS['WMO'][np.where(RAOB_STATIONS['WMO'] == int(station))[0]].values[0]
                        name_idx = 'WMO'
                    except:
                        raise ValueError(
                            f'ICAO or WMO identifier not found, please make sure you provided the correct RAOB ID. If you think this is an error' +
                            'contact the author: https://kylejgillett.github.io/sounderpy/about.html#about-the-author')
                        pass
                    pass

                # begin loading data
                # create dict of data
                new_keys = ['p', 'z', 'T', 'Td', 'u', 'v']
                old_keys = ['pressure', 'height', 'temperature', 'dewpoint', 'u_wind', 'v_wind']  # 'latitude', 'longitude']
                units_list = ['hPa', 'meter', 'degC', 'degC', 'kt', 'kt']
                clean_data = {}
                non_dups = np.concatenate(([True], np.diff(df.to_dict('list')['pressure']) != 0))
                for old_key, new_key, unit in zip(old_keys, new_keys, units_list):
                    clean_data[new_key] = np.array(df.to_dict('list')[old_key])[non_dups] * units(unit)
                clean_data['site_info'] = {
                    'site-id': RAOB_STATIONS[RAOB_STATIONS[name_idx] == station][name_idx].values[0],
                    'site-name': RAOB_STATIONS[RAOB_STATIONS[name_idx] == station]['NAME'].values[0],
                    'site-lctn': RAOB_STATIONS[RAOB_STATIONS[name_idx] == station]['LOC'].values[0],
                    'site-latlon': get_latlon('raob', str(station)),
                    'site-elv': RAOB_STATIONS[RAOB_STATIONS[name_idx] == station]['EL(m)'].values[0],
                    'source': 'RAOB OBSERVED PROFILE',
                    'model': 'no-model',
                    'fcst-hour': 'no-fcst-hour',
                    'run-time': ['none', 'none', 'none', 'none'],
                    'valid-time': [year, month, day, hour]}

                try:
                    # trim data to 98hPa and below for less process time
                    slc = (len(clean_data['p']) - np.where(clean_data['p'] <= 98. * units('hPa'))[0][0])
                    for key in new_keys:
                        clean_data[key] = clean_data[key][:-slc]
                except:
                    pass
        else:
            raise ValueError(
                f'Wyoming Upper Air Archive connection failed -- ensure you have the correct dates and corresponding station identifier\n' +
                f'There is likely no available data for station {station} on {month}/{day}/{year} at {hour}z')
    #########################################################################################################







    ### IGRAv2 OBSERVATIONS ###
    #########################################################################################################
    elif search_for == 'igra':
        for i in range(1, 3):
            try:
                # try siphon IGRA request
                df = IGRAUpperAir.request_data(datetime(int(year), int(month), int(day), int(hour)), station)
                got_data = True
                if got_data:
                    print(f'    > PROFILE FOUND: {station} on {month}/{day}/{year} at {hour}z | From IGRAv2')
                    break
            except:
                got_data = False
                pass

        # if data is found, parse data and create a dict of clean data
        if got_data:
            if clean_it:
                station = \
                IGRA_STATIONS['ID'][np.where(IGRA_STATIONS['ID'].str.contains(station, na=False, case=True))[0]].values[
                    0].strip()

                # create dict of data
                head = df[1]
                df = df[0]
                new_keys = ['p', 'z', 'T', 'Td', 'u', 'v']
                old_keys = ['pressure', 'height', 'temperature', 'dewpoint', 'u_wind', 'v_wind']  # 'latitude', 'longitude']
                units_list = ['hPa', 'meter', 'degC', 'degC', 'kt', 'kt']
                clean_data = {}
                zflag = np.array(df['zflag'])
                pflag = np.array(df['pflag'])
                tflag = np.array(df['tflag'])
                for old_key, new_key, unit in zip(old_keys, new_keys, units_list):
                    clean_data[new_key] = np.array(df.to_dict('list')[old_key])[zflag + pflag + tflag >= 4] * units(unit)
                clean_data['site_info'] = {
                    'site-id': IGRA_STATIONS[IGRA_STATIONS['ID'] == station]['ID'].str.strip().values[0],
                    'site-name': IGRA_STATIONS[IGRA_STATIONS['ID'] == station]['NAME'].str.strip().values[0],
                    'site-lctn': '',
                    'site-latlon': get_latlon('igra', station),
                    'site-elv': IGRA_STATIONS[IGRA_STATIONS['ID'] == station]['EL(m)'].values[0],
                    'source': 'RAOB OBSERVED PROFILE',
                    'model': 'no-model',
                    'fcst-hour': 'no-fcst-hour',
                    'run-time': ['none', 'none', 'none', 'none'],
                    'valid-time': [year, month, day, hour]}
                # correct u & v units
                clean_data['u'] = clean_data['u'] * 1.94384
                clean_data['v'] = clean_data['v'] * 1.94384

        else:
            raise ValueError(
                f'IGRAv2 Dataset connection failed -- ensure you have the correct dates and corresponding station identifier\n' +
                f'There is likely no available data for station {station} on {month}/{day}/{year} at {hour}z')
    #########################################################################################################




    if clean_it:

        ### CHECK AND FIX BAD HEIGHT DATA ###
        ######################################
        def are_heights_bad(arr):
            # check if heights are bad

            # inc flag
            increasing = False

            for i in range(1, len(arr)):
                if arr[i] > arr[i - 1]:
                    increasing = True
                elif increasing:  # are_heights_bad == True
                    return True

            return False


        def fix_bad_heights(heights):
            # fix bad heights

            increasing = False

            last_increasing_index = 0
            # search through height array to find the last
            # value in the sequence to 'increase' correctly
            for i in range(1, len(heights)):
                if heights[i] > heights[i - 1]:  # do so by checking if i is > than i-1
                    # record its index
                    last_increasing_index = i
                else:
                    # the last increasing value is found
                    # return its index
                    break
            last_increasing_value = heights[last_increasing_index]

            # for every 'bad' value afterward, add the last_increasing_index value
            # to it, to correct the height data
            for i in range(last_increasing_index + 1, len(heights)):
                heights[i] += last_increasing_value

                fixed_heights = heights

            return fixed_heights


        bad_heights = are_heights_bad(clean_data['z'])

        if bad_heights:
            clean_data['z'] = fix_bad_heights(clean_data['z'])



        # add plot title to clean_data dict
        clean_data['titles'] = {
            # title specific to RAOB Obs
            'top_title': "RAOB OBSERVED VERTICAL PROFILE",
            'left_title': f"VALID: {clean_data['site_info']['valid-time'][1]}-{clean_data['site_info']['valid-time'][2]}-{clean_data['site_info']['valid-time'][0]} {clean_data['site_info']['valid-time'][3]}Z",
            'right_title': f"{clean_data['site_info']['site-id']} - {clean_data['site_info']['site-name']}, {clean_data['site_info']['site-lctn']} | {clean_data['site_info']['site-latlon'][0]}, {clean_data['site_info']['site-latlon'][1]}    "
        }




        print('    > COMPLETE --------')
        elapsed_time = time.time() - st
        print('    > RUNTIME:', time.strftime("%H:%M:%S", time.gmtime(elapsed_time)))

        if not hush:
            print(
                f"    > SUMMARY: {clean_data['site_info']['valid-time'][3]}Z Launch for {clean_data['site_info']['site-id']}, {clean_data['site_info']['site-name']} at {clean_data['site_info']['valid-time'][1]}"
                f"-{clean_data['site_info']['valid-time'][2]}-{clean_data['site_info']['valid-time'][0]}-{clean_data['site_info']['valid-time'][3]}Z")

            warnings.filterwarnings("ignore")

            sounding_params(clean_data).print_vals()

        return clean_data


    else:
        # if user sets `clean_it=False` return raw data ('dirty') format
        return df

    #########################################################################################################


# __init__.py
from .sounderpy import get_model_data 
from .sounderpy import get_obs_data 
from .sounderpy import get_bufkit_data 
from .sounderpy import acars_data
from .sounderpy import build_sounding
from .sounderpy import build_hodograph
from .sounderpy import build_vad_hodograph
from .sounderpy import build_composite
from .sounderpy import print_variables
from .utils import interp_data
from .utils import find_nearest
from .utils import get_sfc_index
from .utils import make_sfc_based
from .utils import make_sfc_based_3D
from .utils import get_latlon
from .utils import to_file
from .calc import sounding_params
from .calc import vad_params






"""
    SOUNDERPY SOUNDING PARAMETER TIMESERIES

    Purpose of module:

    retrieve sounding data, calculate sounding parameters and plot
    sounding parameters as a timeseries

    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024
"""

####################################################
# CREATE FIGURE
####################################################
fig, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True, figsize=(20, 15))
####################################################



"""
timeseries_data['times']

"""


# AXIS
####################################################
ax1.plot(timeseries_data['times'], timeseries_data['param'], color='cornflowerblue', ls='-', lw=3, label='SRV')
ax1.set_ylabel('0-500m Streamwise\nVorticity (/sec)', color='cornflowerblue', weight='bold', fontsize=18)
ax1.tick_params(axis='y', labelcolor='cornflowerblue', labelsize=15)
ax1.set_ylim(0, 0.045)
ax1.margins(0)

ax1b = ax1.twinx()
ax1b.plot(times, [val.m for val in srw], 'b-', lw=3, label='SRW')
ax1b.set_ylabel('0-500m Storm\nRelative Wind (RM) (kts)', color='b', weight='bold', fontsize=18)
ax1b.tick_params(axis='y', labelcolor='b', labelsize=15)
ax1b.set_ylim(10, 45)
ax1b.margins(0)
####################################################



# KINEMATICS AXIS
####################################################
ax1.plot(times, srv, color='cornflowerblue', ls='-', lw=3, label='SRV')
ax1.set_ylabel('0-500m Streamwise\nVorticity (/sec)', color='cornflowerblue', weight='bold', fontsize=18)
ax1.tick_params(axis='y', labelcolor='cornflowerblue', labelsize=15)
ax1.set_ylim(0, 0.045)
ax1.margins(0)

ax1b = ax1.twinx()
ax1b.plot(times, [val.m for val in srw], 'b-', lw=3, label='SRW')
ax1b.set_ylabel('0-500m Storm\nRelative Wind (RM) (kts)', color='b', weight='bold', fontsize=18)
ax1b.tick_params(axis='y', labelcolor='b', labelsize=15)
ax1b.set_ylim(10, 45)
ax1b.margins(0)
####################################################
















